"""Utility functions used for entropy calculations.

For computational stability, we define two functions and use one of them based on the
applicable setting.
"""

import numpy as np
from numpy.typing import ArrayLike, NDArray


def binary_entropy(e: ArrayLike) -> NDArray[np.float64]:
    r"""Binary entropy function.

    $h(e):\mathbb{R} \rightarrow [0,1]$ is the binary entropy function.

    The domain of the function has been extended such that the optimizer can
    do a proper constrained non-linear optimization.

    Args:
        e: probabilities

    Returns:
        Shannon entropy
    """
    e_arr = np.atleast_1d(e)

    valid_ind = (e_arr > 0) & (e_arr < 1)
    boundary_ind = (e_arr == 0) | (e_arr == 1)
    valid_e = e_arr[valid_ind]

    # Domain extension to help optimizer
    h_e = np.zeros_like(e_arr)
    h_e[valid_ind] = np.abs(
        -valid_e * np.log2(np.abs(valid_e))
        - (1.0 - valid_e) * np.log2(np.abs(1.0 - valid_e))
    )
    h_e[boundary_ind] = 0
    return h_e


def one_minus_binary_entropy(e: ArrayLike) -> NDArray[np.float64]:
    r"""One minus the binary entropy function.

    $h(e):\mathbb{R} \rightarrow [0,1]$ is the binary entropy function.

    The domain of the function has been extended such that the optimizer can
    do a proper constrained non-linear optimization.

    This computation is more precise if $h(e)$ is close to zero or one.
    """
    e_arr = np.atleast_1d(e)

    valid_ind = (e_arr > 0) & (e_arr < 1)
    boundary_ind = (e_arr == 0) | (e_arr == 1)
    valid_e = e_arr[valid_ind]

    # Domain extension to help optimizer
    h_e = np.zeros_like(e_arr)
    h_e[valid_ind] = np.abs(
        valid_e * np.log2(2 * valid_e) + (1 - valid_e) * np.log2(2 - 2 * valid_e)
    )
    h_e[boundary_ind] = 1
    return h_e



"""Classes to perform a Cascade error correction protocol.

The Cascade error correction protocol can be used to correct errors in sifted bit strings. Errors
are detected by calculating the parity between bitstrings of the two parties. The protocol first
divides the messages in different blocks sizes and repeats this a number of passes times where the
block size is doubled each pass. An error is detected and corrected when the parity for a block is
odd. In case the parity for a block is even, it is still possible that (an even number of) errors
exist. These errors can be detected in the next pass, when the block size is doubled and the message
is shuffled. Given enough passes, all errors are expected to be corrected. However, the required
communication of this protocol is high.


The choice of block size, depending on error-rate is based on the following thesis:
Calver, Timothy I., "An Empirical Analysis of the Cascade Secret Key Reconciliation Protocol for Quantum
Key Distribution" (2011). Theses and Dissertations. 1521. (https://scholar.afit.edu/etd/1521/)


Typical usage example:

    >>> from tno.quantum.communication.qkd_key_rate.classical import (
    ...     Message,
    ...     ParityStrategy,
    ...     Permutations,
    ... )
    >>> from tno.quantum.communication.qkd_key_rate.classical.cascade import (
    ...     CascadeCorrector,
    ...     CascadeReceiver,
    ...     CascadeSender,
    ... )
    >>>
    >>> message_length = 100_000
    >>> error_rate = 0.05
    >>> input_message = Message.random_message(message_length=message_length)
    >>> error_message = input_message.apply_errors(error_rate=0.05)
    >>>
    >>> number_of_passes = 8
    >>> sampling_fraction = 0.34
    >>> parity_strategy = ParityStrategy(
    ...     error_rate=error_rate,
    ...     sampling_fraction=sampling_fraction,
    ...     number_of_passes=number_of_passes,
    ... )
    >>>
    >>> permutations = Permutations.random_permutation(
    ...     number_of_passes=number_of_passes, message_size=message_length, random_state=42
    ... )
    >>> alice = CascadeSender(message=input_message, permutations=permutations)
    >>> bob = CascadeReceiver(
    ...     message=error_message,
    ...     permutations=permutations,
    ...     parity_strategy=parity_strategy,
    ... )
    >>> corrector = CascadeCorrector(alice=alice, bob=bob)
    >>>
    >>> print(corrector.correct_errors())  # doctest: +SKIP
    Corrector summary:
    input_alice (Message):   00010010000111101110111110001000001110111101000110...
    output_alice (Message):  00010010000111101110111110001000001110111101000110...
    input_bob (Message):     10010010001111111110111110001000000110111101000100...
    output_bob (Message):    00010010000111101110111110001000001110111101000110...
    input_error (float):     0.05065
    output_error (float):    0.0
    output_length (int):     100000
    number_of_exposed_bits (int):    62301
    key_reconciliation_rate (float):         0.37699
    number_of_communication_rounds (int):    147.0
    number_of_passes (int):  8
    switch_after_pass (int):         8
    sampling_fraction (float):       0.34
"""  # noqa: E501

from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

import numpy as np

from tno.quantum.communication.qkd_key_rate.classical import (
    Corrector,
    CorrectorOutputBase,
    ParityStrategy,
    Permutations,
    ReceiverBase,
    SenderBase,
)

if TYPE_CHECKING:
    from tno.quantum.communication.qkd_key_rate.classical import Message


class CascadeSender(SenderBase):
    """This class encodes all functions available to Cascade sender."""

    def __init__(
        self,
        message: Message,
        permutations: Permutations,
        name: str | None = None,
    ) -> None:
        """Init of CascadeSender.

        Args:
            message: Input message of the sender party
            permutations: List containing permutations for each Cascade pass
            name: Name of the sender party
        """
        super().__init__(message=message, permutations=permutations, name=name)

        self._number_of_exposed_bits = 0
        self._net_exposed_bits = 0
        self.max_exposed_bits = 0
        self.min_exposed_bits = 0

        # Offset for permutation exchange
        self.maximum_number_of_communication_rounds = 1

        self.number_of_passes = permutations.number_of_passes

        self.transcript += "Permutations for bit string shared between alice and bob\n"

    def get_parity(self, index_start: int, index_end: int, pass_number: int) -> int:
        """Get the parity of a specific message block.

        Computation takes into account the permutation applied in that specific pass.

        Args:
            index_start: Start index of the message
            index_end: End index of the message
            pass_number: Cascade pass number

        Returns:
            Parity of substring
        """
        number_of_ones = 0
        for i in range(index_start, index_end):
            if self.message[self.permutations[pass_number][i]] == 1:
                number_of_ones += 1
        self._number_of_exposed_bits += 1
        self._net_exposed_bits += 1

        return number_of_ones % 2


class CascadeReceiver(CascadeSender, ReceiverBase):
    """This class encodes all functions only available to the receiver.

    The receiver is assumed to have a string with errors and is thus assumed to
    correct the errors.
    """

    def __init__(
        self,
        message: Message,
        permutations: Permutations,
        parity_strategy: ParityStrategy,
        name: str | None = None,
    ) -> None:
        """Init of CascadeReceiver.

        Args:
            message: Input message of the sender party
            permutations: List containing permutations for each Cascade pass
            parity_strategy: The cascade parity strategy.
            name: Name of the receiver party

        Raises:
            ValueError: If parity strategy incompatible with number of passes.
        """
        super().__init__(
            message,
            name=name,
            permutations=permutations,
        )
        self.parity_strategy = parity_strategy
        if self.parity_strategy.number_of_passes != self.number_of_passes:
            error_msg = "Parity strategy not compatible with number of passes."
            raise ValueError(error_msg)

        self.block_sizes: list[int] = [0 for _ in range(self.number_of_passes)]
        self.errors_found: list[list[int]] = [[] for _ in range(self.number_of_passes)]

    def correct_errors(self, alice: SenderBase) -> None:
        """This is the main routine.

        Errors in the strings alice and bob are found and corrected

        An upper bound on the number of communications is estimated. This is
        however a loose upper bound. The cascade-routine makes it difficult to
        give tighter bounds

        Args:
            alice: The sending party to correct errors with

        Raises:
            TypeError: If alice is not a cascade sender.
        """
        if not isinstance(alice, CascadeSender):
            error_msg = "Alice is not a cascade sender."
            raise TypeError(error_msg)

        size_blocks_parities = self.parity_strategy.calculate_message_parity_strategy(
            self.message.length
        )

        for index_pass in range(self.number_of_passes):
            # Get the block size and the number of blocks per pass
            (
                block_size,
                number_of_blocks,
            ) = size_blocks_parities[index_pass]
            self.transcript += (
                f"alice sends her parity string for pass {index_pass} to bob.\n"
            )

            self.block_sizes[index_pass] = block_size
            # One round is for sharing the bits from sender to receiver. The other round
            # is to indicate which blocks should be checked from receiver to sender.
            # This number is an upper bound, and in practice might be much lower.
            self.maximum_number_of_communication_rounds += 2 * (
                np.sum(np.ceil(np.log2(self.block_sizes[: index_pass + 1])))
            )

            self.transcript += (
                "bob determines which block parities disagree and corrects them.\n"
            )
            self.transcript += "Error found in (pass, block):\n"

            for index_block in range(number_of_blocks):
                # Check if the parity strings match and otherwise Cascade back
                # to locate the error is this and previous passes.
                self.do_cascade(
                    index_pass, index_block, block_size, alice, size_blocks_parities
                )
            if len(self.errors_found[index_pass]) == 0:
                # In this case we do not have to do all subsequent communication rounds
                self.maximum_number_of_communication_rounds -= 2 * (
                    np.sum(np.ceil(np.log2(self.block_sizes[:index_pass])))
                )

        self.transcript += (
            "Due to the implementation, it is hard to return"
            " the combined messages shared between the parties."
        )

    def check_match_of_parities(
        self,
        alice: CascadeSender,
        current_pass: int,
        current_block: int,
        block_size: int,
    ) -> bool:
        """Checks if the parity strings of alice and bob match.

        This requires communication between the two parties. The parities of
        multiple blocks can be combined in a single message.

        Args:
            alice: The sending party
            current_pass: Index of current pass
            current_block: Index of current block
            block_size: size of the block.

        Returns:
            Boolean if the parity strings of alice and bob match
        """
        index_start = current_block * block_size
        index_finish = min(index_start + block_size, self.message.length)
        parity_alice = alice.get_parity(
            index_start, index_finish, pass_number=current_pass
        )
        parity_bob = self.get_parity(
            index_start, index_finish, pass_number=current_pass
        )
        return bool(parity_alice == parity_bob)

    def do_cascade(
        self,
        current_pass: int,
        current_block: int,
        block_size: int,
        alice: CascadeSender,
        size_blocks_parities: tuple[tuple[int, int], ...],
    ) -> None:
        """Apply the Cascade error correction technique.

        This routine corrects errors in previous passes that become apparent later.
        It is recursively used as long as new errors are found.

        Args:
            current_pass: Index of current pass
            current_block: Index of current block
            block_size: Size of current block
            alice: The sending party
            size_blocks_parities: For each pass, size of the block and number of blocks
        """
        if self.check_match_of_parities(alice, current_pass, current_block, block_size):
            return

        # An error is found, because the parities do not match
        self.transcript += f"({current_pass}, {current_block}), "

        currently_found_error = self.get_error_index(
            current_block, current_pass, block_size, alice
        )
        self.correct_individual_error(currently_found_error)
        self.errors_found[current_pass].append(currently_found_error)

        # Check all preceding iterations if there now is an error to be corrected
        for index_pass in range(current_pass):
            (block_size, _) = size_blocks_parities[index_pass]
            index_block = self.get_block_index(
                currently_found_error, block_size, index_pass
            )
            self.max_exposed_bits += 1
            self.min_exposed_bits += 1
            self.do_cascade(
                index_pass, index_block, block_size, alice, size_blocks_parities
            )

    def get_block_index(self, index: int, block_size: int, index_pass: int) -> int:
        """Returns the block index corresponding to a certain index in a certain pass.

        Args:
            index: Index of a certain bit
            block_size: Size of current block
            index_pass: Index of the current pass

        Returns:
            The block index corresponding to a certain bit index in a certain pass
        """
        original_index = self.permutations.inverted_permutations[index_pass][index]
        return original_index // block_size

    def get_error_index(
        self, index_block: int, index_pass: int, block_size: int, alice: CascadeSender
    ) -> int:
        """Recursively checks the parity of half of the block of both parties.

        Args:
            index_block: Index of the block in which we expect an error
            index_pass: Index of the current pass
            block_size: Size of current block
            alice: The sending party

        Returns:
            The position index of an error
        """
        index_start = index_block * block_size

        # Determine index_finish, either index_start + block_size or message_size
        index_finish = min(self.message.length, index_start + block_size)

        # If the length of the considered part is not a power of 2, the number
        # of exposed bits can vary by one.
        self.max_exposed_bits += np.ceil(np.log2(index_finish - index_start))
        self.min_exposed_bits += np.floor(np.log2(index_finish - index_start))

        while True:
            if (index_finish - index_start) == 1:
                # If we have only a single bit, return that index
                return self.permutations[index_pass][index_start]
            if (index_finish - index_start) == 2:
                # If we have two bits, we check if the first bits agree
                # and return the index of the error accordingly

                parity_alice = alice.get_parity(
                    index_start, index_finish - 1, pass_number=index_pass
                )
                parity_bob = self.get_parity(
                    index_start, index_finish - 1, pass_number=index_pass
                )
                if parity_alice == parity_bob:
                    return self.permutations[index_pass][index_finish - 1]
                return self.permutations[index_pass][index_start]

            # Otherwise, the string contains at least 3 bits. Compute the parity
            # over half of the string
            parity_alice = alice.get_parity(
                index_start,
                index_start + (index_finish - index_start) // 2,
                pass_number=index_pass,
            )
            parity_bob = self.get_parity(
                index_start,
                index_start + (index_finish - index_start) // 2,
                pass_number=index_pass,
            )
            # If the parities match, the error was in the other half of the message
            if parity_alice == parity_bob:
                index_start = index_start + int(
                    np.floor((index_finish - index_start) / 2)
                )
            else:
                index_finish = index_start + int(
                    np.floor((index_finish - index_start) / 2)
                )

    def get_error_rate(self) -> float:
        """Gives the error rate, based on the found errors."""
        index_errors_found = [
            index_error
            for errors_in_pass in self.errors_found
            for index_error in errors_in_pass
        ]

        number_of_errors = len(index_errors_found)
        return number_of_errors / self.message.length

    def get_prior_error_rate(
        self,
        alice: CascadeSender,
        index_start: int = 0,
        index_finish: int | None = None,
    ) -> float:
        """Determine the initial error rate.

        This function is mainly for debugging purposes. Usually,
        the considered bits are private and this value cannot be
        computed.

        Args:
            alice: The sending party
            index_start: Start index of the message
            index_finish: End index of the message

        Returns:
            Initial error rate
        """
        if index_finish is None:
            index_finish = self.message.length
        number_of_differences = 0
        for i in range(index_start, index_finish):
            if alice.message[i] != self.message[i]:
                number_of_differences += 1
        return number_of_differences / self.message.length


@dataclass
class CascadeCorrectorOutput(CorrectorOutputBase):
    """Data class for Cascade Corrector output."""

    number_of_passes: int
    switch_after_pass: int
    sampling_fraction: float


class CascadeCorrector(Corrector):
    """Cascade corrector."""

    def __init__(self, alice: CascadeSender, bob: CascadeReceiver) -> None:
        """Init Cascade corrector.

        Args:
            alice: cascade sender.
            bob: cascade receiver.
        """
        self.alice: CascadeSender = alice
        self.bob: CascadeReceiver = bob

        if self.alice.permutations != self.bob.permutations:
            error_msg = "Permutations Alice not same as permutations Bob."
            raise ValueError(error_msg)

    def summary(self) -> CascadeCorrectorOutput:
        """Calculate a summary object for the error correction.

        Summary contains:

            - original message
            - corrected message
            - error rate (before and after correction)
            - number_of_exposed_bits
            - key_reconciliation_rate
            - protocol specific parameters
        """
        return CascadeCorrectorOutput(
            input_alice=self.alice.original_message,
            output_alice=self.alice.message,
            input_bob=self.bob.original_message,
            output_bob=self.bob.message,
            input_error=self.calculate_error_rate(
                self.alice.original_message, self.bob.original_message
            ),
            output_error=self.calculate_error_rate(
                self.alice.message, self.bob.message
            ),
            output_length=self.alice.message.length,
            number_of_exposed_bits=self.bob.net_exposed_bits,
            key_reconciliation_rate=self.calculate_key_reconciliation_rate(
                exposed_bits=True
            ),
            number_of_communication_rounds=self.bob.maximum_number_of_communication_rounds,
            number_of_passes=self.alice.permutations.number_of_passes,
            switch_after_pass=self.bob.parity_strategy.number_of_passes,
            sampling_fraction=self.bob.parity_strategy.sampling_fraction,
        )



"""Base class for Permutations object."""

from __future__ import annotations

import numpy as np
from numpy.random import RandomState

from tno.quantum.utils.validation import check_random_state


class Permutations:
    """Permutations object."""

    def __init__(self, permutations: list[list[int]]) -> None:
        """Init Permutations.

        Args:
            permutations: list of permutations
        """
        self.permutations = permutations
        self.inverted_permutations = self.calculate_inverted_permutations(permutations)
        self.number_of_passes = len(permutations)

    def __add__(self, other: Permutations) -> Permutations:
        """Combine two permutations."""
        return Permutations(self.permutations + other.permutations)

    def __len__(self) -> int:
        """Return number of passes."""
        return self.number_of_passes

    def __eq__(self, other: object) -> bool:
        """Permutation equality."""
        if not isinstance(other, Permutations):
            return NotImplemented
        return self.permutations == other.permutations

    def __hash__(self) -> int:
        return hash(str(self.permutations))

    def __getitem__(self, pass_number: int) -> list[int]:
        """Return permutation for specific pass.

        Args:
            pass_number: index of the pass number

        Returns:
            The permutation corresponding to pass number.
        """
        return self.permutations[pass_number]

    def shorten_pass(self, pass_idx: int, max_length: int) -> None:
        """Shorten message length.

        In some protocols, the message length decreases, this function
        adjusts the permutations accordingly, by discarding large indices.

        Args:
            pass_idx: Index of the permutation to shorten
            max_length: New message length, and maximum of the entries in
            the permutation (value itself not included)
        """
        self.permutations[pass_idx] = [
            i for i in self.permutations[pass_idx] if i < max_length
        ]
        self.inverted_permutations[pass_idx] = list(
            np.argsort(self.permutations[pass_idx])
        )

    @staticmethod
    def calculate_inverted_permutations(
        permutations: list[list[int]],
    ) -> list[list[int]]:
        """Invert every permutation in a list of permutations."""
        return [list(np.argsort(permutation)) for permutation in permutations]

    @classmethod
    def random_permutation(
        cls,
        number_of_passes: int,
        message_size: int,
        random_state: int | RandomState | None = None,
    ) -> Permutations:
        """Generate a random Permutations object.

        Args:
            number_of_passes: Total number of iterations
            message_size: Size of the message for which to calculate the parity strategy
            random_state: Random state for reproducibility. Defaults to ``None``.
        """
        random_state = check_random_state(random_state, "random_state")
        permutations = [
            random_state.permutation(message_size).tolist()
            for _ in range(number_of_passes)
        ]
        return cls(permutations)



"""Tests for BB84 protocols."""

from copy import deepcopy

import numpy as np

from tno.quantum.communication.qkd_key_rate.quantum import standard_detector
from tno.quantum.communication.qkd_key_rate.quantum.bb84 import (  # type: ignore[attr-defined]
    BB84AsymptoticKeyRateEstimate,
    BB84FiniteKeyRateEstimate,
    BB84FullyAsymptoticKeyRateEstimate,
)


def test_fully_asymptotic() -> None:
    """Test BB84 Fully Asymptotic protocol"""
    detector = deepcopy(standard_detector)

    # All considered losses
    distance = np.arange(1, 100)
    attenuation_factor = 0.1
    attenuation = attenuation_factor * distance

    key_rate: list[float] = [0 for _ in range(len(attenuation))]

    x0 = None
    bounds = None

    fully_asymptotic_key_rate = BB84FullyAsymptoticKeyRateEstimate(detector=detector)
    for i, att in enumerate(attenuation):
        x, rate = fully_asymptotic_key_rate.optimize_rate(
            x0=x0, bounds=bounds, attenuation=float(att)
        )
        key_rate[i] = rate
        x0 = x["mu"]

    assert all(rate > 0 for rate in key_rate)
    assert all(key_rate[i] > key_rate[i + 1] for i in range(len(attenuation) - 1))


def test_asymptotic() -> None:
    """Test BB84 Asymptotic protocol"""
    detector = deepcopy(standard_detector)

    # All considered losses
    distance = np.arange(1, 8)
    attenuation_factor = 0.1
    attenuation = attenuation_factor * distance

    key_rate: list[float] = [0 for _ in range(len(attenuation))]

    number_of_decoy = 3
    asymptotic_key_rate = BB84AsymptoticKeyRateEstimate(
        detector=detector, number_of_decoy=number_of_decoy
    )
    for i, att in enumerate(attenuation):
        _, rate = asymptotic_key_rate.optimize_rate(x0=None, attenuation=float(att))
        key_rate[i] = rate

    assert all(rate > 0 for rate in key_rate)
    assert all(key_rate[i] > key_rate[i + 1] for i in range(len(attenuation) - 1))


def test_finite_bb84() -> None:
    detector = deepcopy(standard_detector)

    # All considered losses
    distance = np.arange(1, 5)
    attenuation_factor = 1
    attenuation = attenuation_factor * distance

    # The detector used at Bob's side
    key_rate: list[float] = [0 for _ in range(len(attenuation))]

    x0 = None
    number_of_pulses = int(10e8)
    number_of_decoy = 1
    finite_key_rate = BB84FiniteKeyRateEstimate(
        detector=detector,
        number_of_pulses=number_of_pulses,
        number_of_decoy=number_of_decoy,
    )
    for i, att in enumerate(attenuation):
        x, rate = finite_key_rate.optimize_rate(x0=x0, attenuation=float(att))
        key_rate[i] = rate
        x0 = np.concatenate(
            (x["mu"], x["probability_basis_X"], x["probability_basis_Z"])
        )

    assert all(rate > 0 for rate in key_rate)
    assert all(key_rate[i] > key_rate[i + 1] for i in range(len(attenuation) - 1))



"""
Computes the maximum effective number of useful secure bits.

For this, an estimate on the number of communication rounds is given,
as well as an efficiency of the error correction routines. The secure
number of bits is determined using the finite key-rate estimates.

Here, BBM92 is used, however, it also works for BB84.
"""

import numpy as np
import pytest

from tno.quantum.communication.qkd_key_rate.classical._required_linktime import (
    compute_efficiency,
    compute_estimate_on_communication,
)
from tno.quantum.communication.qkd_key_rate.quantum import standard_detector
from tno.quantum.communication.qkd_key_rate.quantum.bbm92 import (
    BBM92FiniteKeyRateEstimate,
)

detector = standard_detector.customise(
    dark_count_rate=6e-7,
    polarization_drift=0.0707,
    error_detector=5e-3,
    efficiency_detector=0.1,
)

detector = standard_detector


@pytest.mark.parametrize("error_correction_method", ["Winnow", "Cascade"])
@pytest.mark.parametrize("double_channel_loss", [5, 10, 15, 20])
def test_required_linktime(
    error_correction_method: str, double_channel_loss: float
) -> None:
    # We compute the entropy for these number of pulses
    number_of_pulses = np.array(
        [int(factor * 10**exp) for exp in [9, 10, 11] for factor in range(1, 10)]
    )
    detector_bob = detector

    error_estimate = 0.05  # Estimate for the efficiency of error correction
    epsilon_security = 1e-12
    message_length = 10000
    cost_per_communication = np.ceil(np.log2(1 / epsilon_security))

    number_of_communication_rounds = compute_estimate_on_communication(
        error_correction_method, error_estimate, message_length
    )

    ec_efficiency = compute_efficiency(error_correction_method, error_estimate)
    if ec_efficiency > 0.0 and ec_efficiency < 1.0:
        error_correction_factor = 1 / ec_efficiency

    entropy = []
    for num_pulses in number_of_pulses:
        finite_key_rate = BBM92FiniteKeyRateEstimate(
            detector=detector_bob,
            number_of_pulses=num_pulses,
        )

        # Compute optimal key-rate for the given attenuation, detectors and pulses
        _, optimal_rate = finite_key_rate.optimize_rate(attenuation=double_channel_loss)

        # We use only the pulses in the X-basis for the key
        optimal_number_of_pulses = int(
            np.floor(num_pulses * optimal_rate * error_correction_factor)
        )

        # Correct for the classical authentication required
        net_number_of_pulses = (
            optimal_number_of_pulses
            - number_of_communication_rounds * cost_per_communication
        )

        entropy.append(net_number_of_pulses)

    assert entropy[0] < entropy[-1]



"""Test the Detector class."""

import pytest

from tno.quantum.communication.qkd_key_rate.quantum import Detector


def test_detector() -> None:
    """Test creation of detector"""
    detector = Detector(
        name="standard",
        efficiency_detector=0.2,
        jitter_source=0,
        jitter_detector=5.00e-11,
        dead_time=4.50e-08,
        dark_count_frequency=100,
        detection_frequency=100,
        detection_window=1.00e07,
        efficiency_system=1,
    )
    assert detector


def test_missing_required_field() -> None:
    """Test invalid creation of detector due to missing required field"""
    expected_message = "Field 'name' is required field."
    with pytest.raises(ValueError, match=expected_message):
        Detector(
            name=None,  # type: ignore[arg-type]
            efficiency_detector=0.2,
            jitter_source=0,
            jitter_detector=5.00e-11,
            dead_time=4.50e-08,
            dark_count_frequency=100,
            detection_frequency=100,
            detection_window=1.00e07,
            efficiency_system=1,
        )


def test_invalid_interval_field() -> None:
    """Test invalid creation of detector due to either
    invalid interval or invalid detection_frequency"""
    expected_message = (
        "Either the field 'interval' or 'detection_frequency' is required."
    )
    with pytest.raises(ValueError, match=expected_message):
        Detector(
            name="standard",
            efficiency_detector=0.2,
            jitter_source=0,
            jitter_detector=5.00e-11,
            dead_time=4.50e-08,
            dark_count_frequency=100,
            detection_window=1.00e07,
            efficiency_system=1,
        )

    expected_message = "Incompatible interval and detection frequency."
    with pytest.raises(ValueError, match=expected_message):
        # Case where interval != 1 / detection_frequency
        Detector(
            name="standard",
            efficiency_detector=0.2,
            jitter_source=0,
            jitter_detector=5.00e-11,
            dead_time=4.50e-08,
            dark_count_frequency=100,
            detection_window=1.00e07,
            efficiency_system=1,
            interval=0.01,
            detection_frequency=100 + 20,
        )


def test_invalid_dark_count_rate_field() -> None:
    """Test invalid creation of detector due to either
    invalid dark_count_rate or dark_count_frequency"""
    expected_message = (
        "Either the field 'dark_count_rate' or 'dark_count_frequency' is required."
    )
    with pytest.raises(ValueError, match=expected_message):
        Detector(
            name="standard",
            efficiency_detector=0.2,
            jitter_source=0,
            jitter_detector=5.00e-11,
            dead_time=4.50e-08,
            detection_frequency=100,
            detection_window=1.00e07,
            efficiency_system=1,
        )

    expected_message = "Incompatible dark count rate and dark count frequency."
    with pytest.raises(ValueError, match=expected_message):
        # Case where dark_count_frequency != dark_count_rate * detection_frequency
        Detector(
            name="standard",
            efficiency_detector=0.2,
            jitter_source=0,
            jitter_detector=5.00e-11,
            dead_time=4.50e-08,
            detection_frequency=100,
            detection_window=1.00e07,
            efficiency_system=1,
            dark_count_frequency=100,
            dark_count_rate=1 + 0.05,
        )


def test_invalid_efficiency_field() -> None:
    """Test invalid creation of detector due to either
    invalid efficiency_party or efficiency_detector"""
    expected_message = (
        "Either the field 'efficiency_party' or 'efficiency_detector' is required."
    )
    with pytest.raises(ValueError, match=expected_message):
        Detector(
            name="standard",
            jitter_source=0,
            jitter_detector=5.00e-11,
            dead_time=4.50e-08,
            dark_count_frequency=100,
            detection_frequency=100,
            detection_window=1.00e07,
            efficiency_system=1,
        )

    expected_message = "Incompatible efficiency and efficiency detector."
    with pytest.raises(ValueError, match=expected_message):
        # Case where efficiency_party !=  efficiency_system * efficiency_detector
        Detector(
            name="standard",
            jitter_source=0,
            jitter_detector=5.00e-11,
            dead_time=4.50e-08,
            dark_count_frequency=100,
            detection_frequency=100,
            detection_window=1.00e07,
            efficiency_system=1,
            efficiency_detector=0.2,
            efficiency_party=0.2 + 0.1,
        )


def test_from_file() -> None:
    """Test detector from .csv file."""
    path = "src/tno/quantum/communication/qkd_key_rate/test/base/detectors.csv"
    detectors = Detector.from_file(path)
    for detector in detectors:
        assert isinstance(detector, Detector)


def test_customise() -> None:
    """Test customise detector"""
    detector = Detector(
        name="standard",
        efficiency_detector=0.2,
        jitter_source=0,
        jitter_detector=5.00e-11,
        dead_time=4.50e-08,
        dark_count_frequency=100,
        detection_frequency=1.00e07,
        detection_window=5,
        efficiency_system=1,
    )

    new_dark_count_rate = 1e-8
    new_polarization_drift = 0.1
    new_error_detector = 0.5
    new_efficiency_party = 0.5

    assert detector.customise(
        dark_count_rate=new_dark_count_rate,
        polarization_drift=new_polarization_drift,
        error_detector=new_error_detector,
        efficiency_party=1,
    )

    detector = detector.customise(
        dark_count_rate=new_dark_count_rate,
        polarization_drift=new_polarization_drift,
        error_detector=new_error_detector,
        efficiency_party=new_efficiency_party,
    )
    assert detector.dark_count_rate == new_dark_count_rate
    assert detector.polarization_drift == new_polarization_drift
    assert detector.error_detector == new_error_detector
    assert detector.efficiency_party == new_efficiency_party


def test_new_zero_parameter() -> None:
    """Test setting new parameter to zero."""
    detector = Detector(
        name="standard",
        efficiency_detector=0.2,
        jitter_source=0,
        jitter_detector=5.00e-11,
        dead_time=4.50e-08,
        dark_count_frequency=100,
        detection_frequency=1.00e07,
        detection_window=5,
        efficiency_system=1,
    )

    new_dark_count_rate = 0.0
    detector = detector.customise(dark_count_rate=new_dark_count_rate)
    assert detector.dark_count_rate == new_dark_count_rate



"""Base class for Sending party."""

from __future__ import annotations

from copy import deepcopy
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from tno.quantum.communication.qkd_key_rate.classical import Message, Permutations


class SenderBase:
    """This class encodes all functions available to both sender and receiver."""

    transcript: str
    """This keeps track of the exchanged messages."""

    def __init__(
        self,
        message: Message,
        permutations: Permutations,
        name: str | None = None,
    ) -> None:
        """Init of base sender class.

        Args:
            message: Input message of the sender party
            permutations: Permutations for each pass
            name: Name of the sender party
        """
        self._message = message
        self._original_message = deepcopy(message)

        self.permutations = permutations
        self._number_of_exposed_bits = 0
        self._net_exposed_bits = 0

        self.name = name
        self.transcript = ""

    @property
    def original_message(self) -> Message:
        """Returns original uncorrected message."""
        return self._original_message

    @property
    def message(self) -> Message:
        """Returns the (partially corrected) message."""
        return self._message

    @property
    def number_of_exposed_bits(self) -> int:
        """Counter to track number of exposed bits during protocol."""
        return self._number_of_exposed_bits

    @property
    def net_exposed_bits(self) -> int:
        """Counter to track the `net` number of exposed bits.

        During the Winnow protocol some bits are discarded. The net number of exposed
        bits is the number of exposed bits, minus number of discarded bits.

        For the Cascade protocol the net number of exposed bits is the same as the
        number of exposed bits.
        """
        return self._number_of_exposed_bits



"""
Test the privacy amplification protocol.
"""

import hashlib

import pytest

from tno.quantum.communication.qkd_key_rate._utils import (
    one_minus_binary_entropy as one_minus_h,
)
from tno.quantum.communication.qkd_key_rate.classical import Message
from tno.quantum.communication.qkd_key_rate.classical.privacy_amplification import (
    PrivacyAmplification,
)


@pytest.mark.parametrize(
    ("error_rate_basis_x", "error_correction_loss"),
    [
        (0, 0),
        (0, 10),
        (0.5, 0),
        (0.5, 10),
    ],
)
def test_entropy_estimate(
    error_rate_basis_x: float, error_correction_loss: int
) -> None:
    """Test entropy estimate function"""
    observed_pulses_basis_x = 100
    privacy = PrivacyAmplification(observed_pulses_basis_x, error_rate_basis_x)
    assert privacy.get_entropy_estimate(error_correction_loss) == float(
        observed_pulses_basis_x * one_minus_h(error_rate_basis_x)
        - error_correction_loss
    )


def test_correct_hash() -> None:
    """Test do_hash function"""
    message_length, entropy = 100, 80

    message = Message.random_message(message_length=message_length)
    hash_function = hashlib.shake_256()
    hash_function.update(bytes(message))
    expected_hash = hash_function.digest(entropy)

    privacy = PrivacyAmplification(message_length, 0)
    assert privacy.do_hash(message, entropy) == expected_hash


def test_zero_entropy() -> None:
    """Test zero entropy"""
    message_length = 100
    message = Message.random_message(message_length=message_length)
    privacy = PrivacyAmplification(message_length, 0)

    assert privacy.do_hash(message, entropy=0) == bytes(0)


def test_negative_entropy() -> None:
    """Test negative entropy"""
    message_length = 100
    message = Message.random_message(message_length=message_length)
    privacy = PrivacyAmplification(message_length, 0)

    expected_message = "Entropy is smaller than 0. Secure hash cannot be computed"
    with pytest.raises(ValueError, match=expected_message):
        privacy.do_hash(message, entropy=-1)



"""TNO Quantum communication key-rate module.

The :py:mod:`tno.quantum.communication.qkd_key_rate` package provides python code to
compute optimal protocol parameters for different quantum key distribution (QKD)
protocols.

The codebase is based on the following papers:

- Attema et al. - Optimizing the decoy-state BB84 QKD protocol parameters (2021)(https://doi.org/10.1007/s11128-021-03078-0)
- Ma et al. - Quantum key distribution with entangled photon sources (2007)(http://doi.org/10.1103/PhysRevA.76.012307)


The following quantum protocols are supported:

- BB84 protocol (:py:mod:`~tno.quantum.communication.qkd_key_rate.quantum.bb84`),
- BB84 protocol using a single photon source (:py:mod:`~tno.quantum.communication.qkd_key_rate.quantum.bb84_single_photon`),
- BBM92 protocol (:py:mod:`~tno.quantum.communication.qkd_key_rate.quantum.bbm92`).

The following classical error-correction protocols are supported:

- Cascade (:py:mod:`~tno.quantum.communication.qkd_key_rate.classical.cascade`),
- Winnow (:py:mod:`~tno.quantum.communication.qkd_key_rate.classical.winnow`).

The presented code can be used to

- determine optimal parameter settings needed to obtain the maximum key rate,
- correct errors in exchanged sifted keys for the different QKD protocols,
- apply privacy amplification by calculating secure key using hash function.

Usage examples can be found in relevant submodules.
"""  # noqa: E501

__all__: list[str] = []


__version__ = "2.0.1"



"""Test for BBM92 protocol."""

from copy import deepcopy

import numpy as np

from tno.quantum.communication.qkd_key_rate.quantum import standard_detector
from tno.quantum.communication.qkd_key_rate.quantum.bbm92 import (
    BBM92AsymptoticKeyRateEstimate,
    BBM92FiniteKeyRateEstimate,
)


def test_asymptotic() -> None:
    """Test BBM92 Fully Asymptotic protocol"""
    detector = deepcopy(standard_detector)

    distance = np.arange(1, 20)  # Distance between ALice and Bob in km
    attenuation_factor = 0.2  # Channel loss in dB per km
    attenuation = attenuation_factor * distance

    key_rate: list[float] = [0 for _ in range(len(attenuation))]

    detector_alice = detector.customise(
        dark_count_rate=1e-8,
        polarization_drift=0,
        error_detector=0.1,
        efficiency_party=1,
    )
    detector_bob = detector.customise(
        dark_count_rate=1e-8,
        polarization_drift=0,
        error_detector=0.1,
        efficiency_party=1,
    )

    asymptotic_key_rate = BBM92AsymptoticKeyRateEstimate(
        detector=detector_bob, detector_alice=detector_alice
    )

    for i, att in enumerate(attenuation):
        _, rate = asymptotic_key_rate.optimize_rate(attenuation=float(att))
        key_rate[i] = rate

    assert all(rate > 0 for rate in key_rate)
    assert all(key_rate[i] > key_rate[i + 1] for i in range(len(attenuation) - 1))


def test_finite_bbm92() -> None:
    """Test BBM92 Finite protocol"""
    detector = deepcopy(standard_detector)

    distance = np.arange(1, 10)  # Distance between ALice and Bob in km
    attenuation_factor = 0.2  # Channel loss in dB per km
    attenuation = attenuation_factor * distance

    key_rate: list[float] = [0 for _ in range(len(attenuation))]

    detector_alice = detector.customise(
        dark_count_rate=1e-8,
        polarization_drift=0,
        error_detector=0.1,
        efficiency_party=1,
    )
    detector_bob = detector.customise(
        dark_count_rate=1e-8,
        polarization_drift=0,
        error_detector=0.1,
        efficiency_party=1,
    )

    number_of_pulses = int(5e10)
    finite_key_rate = BBM92FiniteKeyRateEstimate(
        detector=detector_bob,
        detector_alice=detector_alice,
        number_of_pulses=number_of_pulses,
    )

    for i, att in enumerate(attenuation):
        _, rate = finite_key_rate.optimize_rate(attenuation=float(att))
        key_rate[i] = rate

    assert all(rate > 0 for rate in key_rate)



"""Classes to perform key error rate estimate for the single photon BB84 QKD protocol.

The analysis is adjusted to using a single photon source.

This code is based on TNO's BB84 key-rate paper (doi: `10.1007/s11128-021-03078-0`_):

.. _10.1007/s11128-021-03078-0: https://doi.org/10.1007/s11128-021-03078-0


This setting is most similar to the originally proposed BB84 protocol, where single
photon quantum states are send. Generating single photon states is hard in practice,
hence the general approach is to use an attunable laser source and use multiple
intensity settings. If we instead use single photon states, much of the analysis
simplifies as we know each pulse is safe against for instance photon-number
splitting (PNS) attacks.

- Asymptotic Key Rate
    The number of pulses is asymptotic and we use only a single intensity setting,
    which we optimize. The other functions are similar to the standard BB84 protocol.

    >>> from tno.quantum.communication.qkd_key_rate.quantum.bb84_single_photon import (
    ...     BB84SingleAsymptoticKeyRateEstimate,
    ... )
    >>> from tno.quantum.communication.qkd_key_rate.quantum import standard_detector
    >>>
    >>> detector = standard_detector.customise(
    ...     dark_count_rate=1e-8,
    ...     polarization_drift=0,
    ...     error_detector=0.1,
    ...     efficiency_party=1,
    ... )
    >>>
    >>> finite_key_rate = BB84SingleAsymptoticKeyRateEstimate(detector=detector)
    >>> x, rate = finite_key_rate.optimize_rate(attenuation=0.2)
    >>> print(f"{x['mu']=}, {rate=}")
    x['mu']=array([0.89916328]), rate=0.858693993504011
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Any

import numpy as np
import scipy
from numpy.typing import ArrayLike, NDArray

from tno.quantum.communication.qkd_key_rate._utils import binary_entropy as h
from tno.quantum.communication.qkd_key_rate.quantum._config import NLP_CONFIG
from tno.quantum.communication.qkd_key_rate.quantum._keyrate import (
    AsymptoticKeyRateEstimate,
)

if TYPE_CHECKING:
    from tno.quantum.communication.qkd_key_rate.quantum import Detector


def compute_gain_and_error_rate(
    detector: Detector, attenuation: float
) -> tuple[float, float]:
    """Computes the total gain and error rate of the channel.

    These computations are straight-forward and have a closed
    expression for single photon sources.

    Args:
        detector: Bob's detector
        attenuation: Attenuation of the channel

    Returns:
        - The gain per intensity. The probability of an event, given a pulse.
        - The error rate per intensity.
    """
    dark_count_rate = detector.dark_count_rate
    efficiency_bob = detector.efficiency_party
    polarization_drift = detector.polarization_drift

    efficiency_channel = np.power(10, -attenuation / 10)
    efficiency = efficiency_bob * efficiency_channel

    gain = 1 - np.power(1 - dark_count_rate, 2) * (1 - efficiency)
    yield_times_error_single = (
        1
        - (1 - dark_count_rate) * efficiency * np.cos(2 * polarization_drift)
        - np.power(1 - dark_count_rate, 2) * (1 - efficiency)
    ) / 2
    error_rate = yield_times_error_single / gain
    return gain, error_rate


class BB84SingleAsymptoticKeyRateEstimate(AsymptoticKeyRateEstimate):
    """Key-rate modules for BB84 when a single photon source is used.

    We assume that all states remain safe and photon number splitting attacks are
    not possible.
    """

    def __init__(self, detector: Detector, **kwargs: Any) -> None:
        """Init of BB84SingleAsymptoticKeyRateEstimate.

        Args:
            detector: The detector used at Bob's side
            kwargs: protocol specific input
        """
        super().__init__(detector=detector, args=kwargs)

    def compute_rate(self, mu: float | ArrayLike, attenuation: float) -> float:  # type: ignore[override]
        """Computes the key-rate given an intensity and an attenuation.

        Args:
            mu: Intensity
            attenuation: Attenuation

        Returns:
            Key-rate
        """
        mu = np.atleast_1d(mu)
        if mu.shape != (1,):
            error_msg = "Multiple intensities were given, only one is expected."
            raise ValueError(error_msg)

        Q, E = compute_gain_and_error_rate(self.detector, attenuation)
        return float(mu[0] * Q * (1 - 2 * h(E)))

    def _extract_parameters(
        self, x: NDArray[np.float64]
    ) -> dict[str, NDArray[np.float64]]:
        return {"mu": x}

    def optimize_rate(
        self,
        *,
        attenuation: float,
        x0: ArrayLike | None = None,
        bounds: list[tuple[float, float]] | None = None,
    ) -> tuple[dict[str, NDArray[np.float64]], float]:
        """Function to optimize the key-rate.

        Args:
            attenuation: Loss in dB for the channel
            x0: Initial search value, default value [0.5] is chosen.
            bounds: Bounds on search range, default [(0.1, 0.9)]

        Returns:
            Optimized intensity and key-rate

        Raises:
            ValueError: When x0 or bounds are given with invalid dimensions.
            ValueError: when the found key-rate is negative.
        """
        if bounds is None:
            # The lower and upper bound on the laser intensity
            lower_bound = np.array([0.1])
            upper_bound = np.array([0.9])
        else:
            if len(bounds) != 1:
                error_msg = (
                    "Invalid dimensions input bounds. Expected 1 upper- and lower"
                    f" bound but received {len(bounds)} bounds."
                )
                raise ValueError(error_msg)
            lower_bound = np.array([bound[0] for bound in bounds])
            upper_bound = np.array([bound[1] for bound in bounds])
        bounds = scipy.optimize.Bounds(lower_bound, upper_bound)

        if x0 is None:
            x0 = [0.5]
        elif len(np.asarray(x0)) != 1:
            error_msg = (
                "Invalid number of inputs. Expected 1 intensity."
                f" Received {len(np.asarray(x0))} intensities."
            )
            raise ValueError(error_msg)

        args = {"attenuation": attenuation}
        res = scipy.optimize.minimize(
            self._f, x0, args=args, bounds=bounds, **NLP_CONFIG
        )

        mu = res.x
        rate = np.atleast_1d(-res.fun)[0]
        if rate < 0:
            error_msg = "Optimization resulted in a negative key rate."
            raise ValueError(error_msg)
        return self._extract_parameters(mu), float(rate)



"""Test the Permutations object."""

from tno.quantum.communication.qkd_key_rate.classical import Permutations


def test_random_permutations_init() -> None:
    """Test creating random permutation"""
    number_of_passes = 2
    message_size = 6

    permutations = Permutations.random_permutation(
        number_of_passes=number_of_passes, message_size=message_size
    )

    assert isinstance(permutations, Permutations)
    assert permutations.number_of_passes == number_of_passes
    for permutation in permutations.permutations:
        assert len(permutation) == message_size


def test_inverted_permutations() -> None:
    """Test inverted permutation"""
    number_of_passes = 2
    message_size = 6

    permutations = Permutations.random_permutation(
        number_of_passes=number_of_passes, message_size=message_size
    )

    for permutation, inverted_permutation in zip(
        permutations.permutations, permutations.inverted_permutations
    ):
        assert [permutation[i] for i in inverted_permutation] == list(
            range(message_size)
        )


def test_addition() -> None:
    """Test combining permutations"""
    p1 = Permutations([[1, 2, 3], [4, 5, 6]])
    p2 = Permutations([[7, 8, 9]])
    p3 = Permutations([[1, 2, 3], [4, 5, 6], [7, 8, 9]])
    assert p1 + p2 == p3

    p1 += p2
    assert p1 == p3


def test_length() -> None:
    """Test length function of permutations"""
    number_of_passes = 10
    p = Permutations.random_permutation(
        number_of_passes=number_of_passes, message_size=100
    )
    assert len(p) == 10


def test_shorten_pass() -> None:
    """Test shortening permutations"""
    permutations = Permutations([list(range(10)), list(range(10))])

    permutations.shorten_pass(0, 7)
    permutations.shorten_pass(1, 4)
    assert permutations[0] == list(range(7))
    assert permutations[1] == list(range(4))
    assert permutations.inverted_permutations[0] == list(range(7))
    assert permutations.inverted_permutations[1] == list(range(4))

    permutations = Permutations.random_permutation(number_of_passes=3, message_size=100)
    permutations.shorten_pass(0, 5)
    permutations.shorten_pass(1, 50)
    permutations.shorten_pass(2, 110)

    for permutation, inverted_permutation in zip(
        permutations.permutations, permutations.inverted_permutations
    ):
        assert [permutation[i] for i in inverted_permutation] == list(
            range(len(permutation))
        )



"""Base class for Message object."""

from __future__ import annotations

from collections.abc import Iterable, Iterator
from dataclasses import dataclass

from numpy.random import RandomState

from tno.quantum.utils.validation import check_binary, check_random_state, check_real


@dataclass(init=False)
class Message:
    """Message object containing binary bits."""

    message: list[int]
    """Message object."""

    def __init__(self, message: Iterable[int | str]) -> None:
        """Init :py:class:`Message`.

        Args:
            message: The message, iterable object with binary items.

        Raises:
            TypeError: If `message` contains items that can't be converted to a binary.
            ValueError: If `message` contains items that can't be converted to a binary.
        """
        self.message = [
            check_binary(value, f"message[{i}]") for i, value in enumerate(message)
        ]

    @property
    def length(self) -> int:
        """Length of message."""
        return len(self.message)

    def __getitem__(self, key: int) -> int:
        """Return value of message for specific index.

        Args:
            key: The index at which the value should be returned.
        """
        return self.message[key]

    def __setitem__(self, key: int, value: int) -> None:
        """Set key of message to specific value.

        Args:
            key: The index at which the value should be set.
            value: The value to be inserted at the specified index.
        """
        self.message[key] = value

    def __bytes__(self) -> bytes:
        """Bytes representation of message."""
        return bytes("".join(str(x) for x in self.message).encode("utf-8"))

    def __str__(self) -> str:
        """String representation of message."""
        res = "".join(str(i) for i in self.message)[:50]
        if self.length > 50:
            res += "..."
        return res

    def pop(self, index: int = -1) -> int:
        """Remove bit at a specific index from message."""
        return self.message.pop(index)

    def apply_permutation(self, permutation: list[int]) -> None:
        """Apply a permutation to the message.

        Args:
            permutation: The permutation that is applied

        Raises:
            ValueError: If message is incompatible with permutation.
        """
        if self.length != len(permutation):
            error_msg = "Message is incompatible with permutation."
            raise ValueError(error_msg)
        self.message = [self.message[i] for i in permutation]

    @classmethod
    def random_message(
        cls, message_length: int, random_state: int | RandomState | None = None
    ) -> Message:
        """Generate a random message.

        Args:
            message_length: Length of random message
            random_state: Random state for reproducibility. Defaults to ``None``.

        Returns:
            random message
        """
        random_state = check_random_state(random_state, "random_state")
        return cls(list(random_state.randint(2, size=message_length)))

    def __iter__(self) -> Iterator[int]:
        """Create iterator for message bits."""
        return iter(self.message)

    def apply_errors(
        self, error_rate: float, random_state: int | RandomState | None = None
    ) -> Message:
        """Apply errors to message.

        Args:
            error_rate: probability that an error occurs.
            random_state: Random state for reproducibility. Defaults to ``None``

        Returns:
            Message to which errors are applied.

        Raises: ValueError if error rate not provided as percentage.
        """
        error_rate = check_real(error_rate, "error_rate", l_bound=0, u_bound=1)
        random_state = check_random_state(random_state, "random_state")

        error_message = [x if random_state.rand() > error_rate else 1 - x for x in self]
        return Message(error_message)



"""Test the Schedule object."""

import pytest

from tno.quantum.communication.qkd_key_rate.classical._schedule import (
    Schedule,
)


@pytest.mark.parametrize(
    ("error_rate", "expected_schedule"),
    [
        (0.009, Schedule([1, 0, 1, 1, 1, 0, 0, 1])),
        (0.019, Schedule([1, 1, 2, 0, 0, 0, 0, 1])),
        (0.029, Schedule([1, 3, 0, 0, 0, 1, 0, 1])),
        (0.039, Schedule([2, 1, 1, 1, 0, 0, 0, 1])),
        (0.049, Schedule([2, 1, 1, 0, 0, 1, 1, 1])),
        (0.059, Schedule([2, 1, 2, 1, 0, 0, 0, 1])),
        (0.069, Schedule([2, 1, 2, 2, 1, 0, 0, 0])),
        (0.079, Schedule([2, 2, 1, 2, 0, 0, 0, 1])),
        (0.089, Schedule([2, 2, 1, 2, 1, 0, 0, 1])),
        (0.099, Schedule([3, 1, 2, 1, 0, 0, 0, 1])),
        (0.109, Schedule([3, 2, 1, 1, 0, 0, 0, 1])),
        (0.119, Schedule([3, 2, 3, 0, 0, 0, 0, 1])),
        (0.129, Schedule([4, 2, 2, 1, 0, 0, 0, 1])),
        (0.139, Schedule([4, 2, 2, 0, 0, 0, 0, 2])),
        (0.149, Schedule([4, 3, 2, 0, 0, 0, 1, 0])),
    ],
)
def test_init_schedule_from_error_rate(
    error_rate: float, expected_schedule: Schedule
) -> None:
    """Test schedule from error rate"""

    schedule = Schedule.schedule_from_error_rate(error_rate=error_rate)
    assert schedule == expected_schedule


def test_stop_high_error() -> None:
    """Test set block schedule from to high error rate"""
    estimated_error_rate = 0.16
    expected_msg = "Error rate too high for secure protocol"
    with pytest.raises(ValueError, match=expected_msg):
        Schedule.schedule_from_error_rate(error_rate=estimated_error_rate)


@pytest.mark.parametrize("error_rate", [0.009, 0.019, 0.069, 0.139, 0.149])
def test_next_pass(error_rate: float) -> None:
    """Test next pass functionality"""
    schedule = Schedule.schedule_from_error_rate(error_rate=error_rate)

    assert schedule.remaining_passes == len(schedule)
    res = []
    for pass_number in range(len(schedule)):
        res.append(schedule.next_pass())
        assert schedule.remaining_passes == len(schedule) - pass_number - 1

    for i in range(8):
        assert res.count(i) == schedule.schedule[i]

    with pytest.raises(IndexError) as _:
        schedule.next_pass()



"""Test the Winnow error correction protocol."""

import numpy as np

from tno.quantum.communication.qkd_key_rate.classical import (
    Message,
    Permutations,
    Schedule,
)
from tno.quantum.communication.qkd_key_rate.classical.winnow import (
    WinnowCorrector,
    WinnowReceiver,
    WinnowSender,
)
from tno.quantum.utils.validation import check_random_state

random_state = check_random_state(None, "random_state")


def test_correctness() -> None:
    """Test correctness for Winnow protocol"""
    message_length = 60000
    error_rate = 0.01
    input_message = Message.random_message(message_length=message_length)
    error_message = input_message.apply_errors(error_rate=error_rate)

    schedule = Schedule.schedule_from_error_rate(error_rate=error_rate)
    number_of_passes = np.sum(schedule.schedule)
    permutations = Permutations.random_permutation(
        number_of_passes=number_of_passes, message_size=message_length
    )

    alice = WinnowSender(
        message=input_message, permutations=permutations, schedule=schedule
    )
    bob = WinnowReceiver(
        message=error_message, permutations=permutations, schedule=schedule
    )

    corrector = WinnowCorrector(alice=alice, bob=bob)
    results = corrector.correct_errors()

    assert results.output_alice == results.output_bob
    assert abs(1 - results.input_error / error_rate) < 0.2  #  noqa: PLR2004
    assert results.output_error == 0
    assert (
        abs(
            (results.output_length - results.number_of_exposed_bits) / message_length
            - results.key_reconciliation_rate
        )
        < 0.001  #  noqa: PLR2004
    )
